"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var rxjs_1 = require("rxjs");
var event_metadata_1 = require("./event-metadata");
/** @PropertyDecoratorFactory */
function EventSource() {
    var args = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        args[_i] = arguments[_i];
    }
    var paramsArg;
    if (args.length > 0) {
        paramsArg = args[0];
    }
    if (!paramsArg || paramsArg instanceof Function) {
        return EventSource.WithParams.apply(EventSource, [undefined].concat(args));
    }
    else {
        return EventSource.WithParams.apply(EventSource, [paramsArg].concat(args.slice(1)));
    }
}
exports.EventSource = EventSource;
(function (EventSource) {
    /** @PropertyDecoratorFactory */
    function WithParams(eventType) {
        var methodDecorators = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            methodDecorators[_i - 1] = arguments[_i];
        }
        /** @PropertyDecorator */
        return function (target, propertyKey) {
            // If an eventType wasn't specified...
            if (!eventType) {
                // Try to deduce the eventType from the propertyKey
                if (propertyKey.endsWith("$")) {
                    eventType = propertyKey.substring(0, propertyKey.length - 1);
                }
                else {
                    throw new Error("@EventSource error: eventType could not be deduced from propertyKey \"" + propertyKey + "\" (only keys ending with '$' can be auto-deduced).");
                }
            }
            // Create the event source metadata for the decorated property
            EventSource.CreateMetadata(eventType, target, propertyKey);
            // Apply any method decorators to the facade function
            methodDecorators.forEach(function (methodDecorator) { return methodDecorator(target, eventType, Object.getOwnPropertyDescriptor(target, eventType)); });
        };
    }
    EventSource.WithParams = WithParams;
    function Bootstrap(targetInstance) {
        // Copy all event metadata from the constructor to the target instance
        var metadataMap = event_metadata_1.EventMetadata.CopyMetadata(event_metadata_1.EventMetadata.GetOwnMetadataMap(targetInstance), event_metadata_1.EventMetadata.CopyInherittedMetadata(targetInstance.constructor), true);
        // Iterate over each of the target properties for each proxied event type used in this class
        metadataMap.forEach(function (propertySubjectMap, eventType) { return propertySubjectMap.forEach(function (subjectInfo, propertyKey) {
            // If the event proxy subject hasn't been created for this property yet...
            if (!subjectInfo.subject) {
                // Create a new Subject
                subjectInfo.subject = new rxjs_1.Subject();
            }
            // Make the observable accessible to the target instance
            targetInstance[propertyKey] = subjectInfo.subject.asObservable();
        }); });
    }
    EventSource.Bootstrap = Bootstrap;
    var Facade;
    (function (Facade) {
        /** @description
         *  Creates an event facade function (the function that is invoked during an event) for the given event type.
         */
        function Create(eventType) {
            return Object.assign(function (value) {
                // Iterate over all class properties that have a proxy subject for this event type and notify each subscriber
                event_metadata_1.EventMetadata.GetPropertySubjectMap(eventType, this).forEach(function (subjectInfo) { return subjectInfo.subject.next(value); });
            }, { eventType: eventType });
        }
        Facade.Create = Create;
    })(Facade = EventSource.Facade || (EventSource.Facade = {}));
    function CreateMetadata(type, target, propertyKey) {
        if (target[type] && target[type].eventType !== type) {
            // Make sure the target class doesn't have a custom method already defined for this event type
            throw new Error("@EventSource metadata creation failed. Class already has a custom " + type + " method.");
        }
        else if (!target[type]) {
            // Assign the facade function for the given event type to the appropriate target class method
            target[type] = Facade.Create(type);
        }
        // Create initial metadata
        var metadata = { subject: undefined };
        // Add the propertyKey to the class' metadata
        event_metadata_1.EventMetadata.GetOwnPropertySubjectMap(type, target.constructor).set(propertyKey, metadata);
        return metadata;
    }
    EventSource.CreateMetadata = CreateMetadata;
})(EventSource = exports.EventSource || (exports.EventSource = {}));
