"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var rxjs_1 = require("rxjs");
var ObservableUtil;
(function (ObservableUtil) {
    /** @description
     *  Creates an observable from the given property.
     */
    function CreateFromProperty(property) {
        if (property instanceof rxjs_1.Subject) {
            return property.asObservable();
        }
        else if (property instanceof rxjs_1.Observable) {
            return property;
        }
        else {
            return rxjs_1.Observable.of(property);
        }
    }
    ObservableUtil.CreateFromProperty = CreateFromProperty;
    /**
     *  @param target The target object.
     *  @param path The target property path.
     *  @description Creates an observable chain from the given property path.
     *
     *  Note: Any property in the path that isn't an Observable or Subject will implicitly be converted to an Observable.
     */
    function CreateFromPropertyPath(target, path) {
        var lastPropertyKey = "target";
        /**
         * @param target The target object.
         * @param propertyKeys The list of property keys in the path.
         * @description
         * Creates an observable chain from the property path and returns the value of the last property in the path.
         **/
        return (function resolveProperty(target, propertyKeys) {
            if (!target) {
                throw new Error("@StateEmitter - Failed to deduce dynamic path \"" + path + "\": " + lastPropertyKey + " is undefined.");
            }
            // Get the property key
            var curPropertyKey = lastPropertyKey = propertyKeys[0];
            // Create an observable from the properties value
            return CreateFromProperty(target[curPropertyKey]).flatMap(function (target) {
                // If it's the last property in the path...
                if (propertyKeys.length === 1) {
                    // Return the value
                    return rxjs_1.Observable.of(target);
                }
                else {
                    // Otherwise, return the next property in the path
                    return resolveProperty(target, propertyKeys.slice(1));
                }
            });
        })(target, path.split("."));
    }
    ObservableUtil.CreateFromPropertyPath = CreateFromPropertyPath;
    /**
     * @param target The target object.
     * @param path The target property path.
     * @description Checks if the given property path is dynamic.
     * A path is considered dynamic or non-static if:
     *
     * - It contains any Observables, or Subjects that are not the terminal property in the path.
     * - It does not terminate with a Subject.
     */
    function IsDynamicPropertyPath(target, path) {
        // Get all property keys in the path
        var propertyKeys = path.split(".");
        return propertyKeys.some(function (propertyKey, index) {
            // Get the property value
            target = target[propertyKey];
            // If this is the last property in the path...
            if (index === propertyKeys.length - 1) {
                // The property must be a Subject to be emittable
                return !(target instanceof rxjs_1.Subject);
            }
            else {
                // The property must not contain any Observables or non-terminal Subjects to be emittable
                return target instanceof rxjs_1.Observable || target instanceof rxjs_1.Subject;
            }
        });
    }
    ObservableUtil.IsDynamicPropertyPath = IsDynamicPropertyPath;
    // NOTE: Static property paths will result in a Subject (two-way binding), while dynamic property paths will result in an Observable (one-way binding)
    function ResolvePropertyPath(target, path) {
        // If the path is dynamic...
        if (IsDynamicPropertyPath(target, path)) {
            // Create an observable chain from the property path
            return CreateFromPropertyPath(target, path);
        }
        else {
            // Resolve the subject from the path
            return ResolveStaticPropertyPath(target, path);
        }
    }
    ObservableUtil.ResolvePropertyPath = ResolvePropertyPath;
    function ResolveStaticPropertyPath(target, path) {
        // Statically access each property and get the terminating subject
        return path.split(".").reduce(function (target, key) { return target[key]; }, target);
    }
    ObservableUtil.ResolveStaticPropertyPath = ResolveStaticPropertyPath;
    /**
     * @param target The target object.
     * @param path The target property path.
     * @param value The new value of the property.
     * @param mergeValue Whether or not the newly emitted value should be merged into the last emitted value.
     * @description Traverses the property path for a Subject, and appropriately wraps and emits the given value from the Subject.
     */
    function UpdateDynamicPropertyPathValue(target, path, value, mergeValue) {
        // Get all property keys in the path
        var propertyKeys = path.split(".");
        var subject;
        var subjectIndex;
        // Iterate over each property key to find a Subject
        propertyKeys.every(function (propertyKey, index) {
            // If the current property is a Subject...
            if (target[propertyKey] instanceof rxjs_1.Subject) {
                // Record this subject
                subject = target[propertyKey];
                subjectIndex = index;
                // Stop searching for a subject
                return false;
            }
            return true;
        });
        // If there is no Subject in the path, throw an error
        if (!subject) {
            throw new Error("Failed to update value for dynamic property " + path + " - Path does not contain a Subject.");
        }
        // Resolve the subject value
        var updatedSubjectValue = propertyKeys
            .slice(subjectIndex + 1)
            .reduceRight(function (value, propertyKey) {
            return (_a = {}, _a[propertyKey] = value, _a);
            var _a;
        }, value);
        if (mergeValue) {
            // Get the last value from the subject and emit the merged properties
            subject.take(1).subscribe(function (lastValue) { return subject.next(Object.assign(lastValue, updatedSubjectValue)); });
        }
        else {
            // Emit the new value
            subject.next(updatedSubjectValue);
        }
    }
    ObservableUtil.UpdateDynamicPropertyPathValue = UpdateDynamicPropertyPathValue;
})(ObservableUtil = exports.ObservableUtil || (exports.ObservableUtil = {}));
